#!/usr/local/bin/python
#
# Exif information decoder
# Written by Thierry Bousch <bousch@topo.math.u-psud.fr>
# Public Domain
#
# $Id: exifdump.py,v 1.13 1999/08/21 14:46:36 bousch Exp $
#
# Since I don't have a copy of the Exif standard, I got most of the
# information from the TIFF/EP draft International Standard:
#
#   ISO/DIS 12234-2
#   Photography - Electronic still picture cameras - Removable Memory
#   Part 2: Image data format - TIFF/EP
#
# You can (and should) get a copy of this document from PIMA's web site;
# their URL is:  http://www.pima.net/it10a.htm
# You'll also find there some documentation on DCF (Digital Camera Format)
# which is based on Exif.
#
# Another must-read is the TIFF 6.0 specification, which can be
# obtained from Adobe's FTP site, at
# ftp://ftp.adobe.com/pub/adobe/devrelations/devtechnotes/pdffiles/tiff6.pdf
#
# Many thanks to Doraneko <doraneko@unioncg.or.jp> for filling the
# holes in the Exif tag list.

import sys

EXIF_TAGS = {
	0x100:	"ImageWidth",
	0x101:	"ImageLength",
	0x102:	"BitsPerSample",
	0x103:	"Compression",
	0x106:	"PhotometricInterpretation",
	0x10A:	"FillOrder",
	0x10D:	"DocumentName",
	0x10E:	"ImageDescription",
	0x10F:	"Make",
	0x110:	"Model",
	0x111:	"StripOffsets",
	0x112:	"Orientation",
	0x115:	"SamplesPerPixel",
	0x116:	"RowsPerStrip",
	0x117:	"StripByteCounts",
	0x11A:	"XResolution",
	0x11B:	"YResolution",
	0x11C:	"PlanarConfiguration",
	0x128:	"ResolutionUnit",
	0x12D:	"TransferFunction",
	0x131:	"Software",
	0x132:	"DateTime",
	0x13B:	"Artist",
	0x13E:	"WhitePoint",
	0x13F:	"PrimaryChromaticities",
	0x156:	"TransferRange",
	0x200:	"JPEGProc",
	0x201:	"JPEGInterchangeFormat",
	0x202:	"JPEGInterchangeFormatLength",
	0x211:	"YCbCrCoefficients",
	0x212:	"YCbCrSubSampling",
	0x213:	"YCbCrPositioning",
	0x214:	"ReferenceBlackWhite",
	0x828D:	"CFARepeatPatternDim",
	0x828E:	"CFAPattern",
	0x828F:	"BatteryLevel",
	0x8298:	"Copyright",
	0x829A:	"ExposureTime",
	0x829D:	"FNumber",
	0x83BB:	"IPTC/NAA",
	0x8769:	"ExifOffset",
	0x8773:	"InterColorProfile",
	0x8822:	"ExposureProgram",
	0x8824:	"SpectralSensitivity",
	0x8825:	"GPSInfo",
	0x8827:	"ISOSpeedRatings",
	0x8828:	"OECF",
	0x9000:	"ExifVersion",
	0x9003:	"DateTimeOriginal",
	0x9004:	"DateTimeDigitized",
	0x9101:	"ComponentsConfiguration",
	0x9102:	"CompressedBitsPerPixel",
	0x9201:	"ShutterSpeedValue",
	0x9202:	"ApertureValue",
	0x9203:	"BrightnessValue",
	0x9204:	"ExposureBiasValue",
	0x9205:	"MaxApertureValue",
	0x9206:	"SubjectDistance",
	0x9207:	"MeteringMode",
	0x9208:	"LightSource",
	0x9209:	"Flash",
	0x920A:	"FocalLength",
	0x927C:	"MakerNote",
	0x9286:	"UserComment",
	0x9290:	"SubSecTime",
	0x9291:	"SubSecTimeOriginal",
	0x9292:	"SubSecTimeDigitized",
	0xA000:	"FlashPixVersion",
	0xA001:	"ColorSpace",
	0xA002:	"ExifImageWidth",
	0xA003:	"ExifImageLength",
	0xA005:	"InteroperabilityOffset",
	0xA20B:	"FlashEnergy",			# 0x920B in TIFF/EP
	0xA20C:	"SpatialFrequencyResponse",	# 0x920C    -  -
	0xA20E:	"FocalPlaneXResolution",	# 0x920E    -  -
	0xA20F:	"FocalPlaneYResolution",	# 0x920F    -  -
	0xA210:	"FocalPlaneResolutionUnit",	# 0x9210    -  -
	0xA214:	"SubjectLocation",		# 0x9214    -  -
	0xA215:	"ExposureIndex",		# 0x9215    -  -
	0xA217:	"SensingMethod",		# 0x9217    -  -
	0xA300:	"FileSource",
	0xA301:	"SceneType",
    # why not merge the two dicts, there's no tag overlap?
	0x1:	"InteroperabilityIndex",
	0x2:	"InteroperabilityVersion",
	0x1000:	"RelatedImageFileFormat",
	0x1001:	"RelatedImageWidth",
	0x1002:	"RelatedImageLength",
}

def s2n_motorola(str):
  x = 0
  for c in str:
    x = (x << 8) | ord(c)
  return x

def s2n_intel(str):
  x = 0
  y = 0
  for c in str:
    x = x | (ord(c) << y)
    y = y + 8
  return x


class Fraction:

  def __init__(self, num, den):
    self.num = num
    self.den = den

  def __repr__(self):
    # String representation
    return '%d/%d' % (self.num, self.den)


class TIFF_file:

  def __init__(self, data):
    self.data = data
    self.endian = data[0]
  
  def s2n(self, offset, length, signed=0):
    slice = self.data[offset:offset+length]
    if self.endian == 'I':
      val = s2n_intel(slice)
    else:
      val = s2n_motorola(slice)
    # Sign extension ?
    if signed:
      msb = 1 << (8*length - 1)
      if val & msb:
        val = val - (msb << 1)
    return val

  def first_IFD(self):
    return self.s2n(4, 4)
  
  def next_IFD(self, ifd):
    entries = self.s2n(ifd, 2)
    return self.s2n(ifd + 2 + 12 * entries, 4)

  def list_IFDs(self):
    i = self.first_IFD()
    a = []
    while i:
      a.append(i)
      i = self.next_IFD(i)
    return a
    
  def dump_IFD(self, ifd):
    entries = self.s2n(ifd, 2)
    a = []
    for i in range(entries):
      entry = ifd + 2 + 12*i
      tag = self.s2n(entry, 2)
      type = self.s2n(entry+2, 2)
      if not 1 <= type <= 10:
        continue # not handled
      typelen = [ 1, 1, 2, 4, 8, 1, 1, 2, 4, 8 ] [type-1]
      count = self.s2n(entry+4, 4)
      offset = entry+8
      if count*typelen > 4:
        offset = self.s2n(offset, 4)
      if type == 2:
        # Special case: nul-terminated ASCII string
        values = self.data[offset:offset+count-1]
      else:
        values = []
        signed = (type == 6 or type >= 8)
        for j in range(count):
          if type % 5:
            # Not a fraction
            value_j = self.s2n(offset, typelen, signed)
          else:
            # The type is either 5 or 10
            value_j = Fraction(self.s2n(offset,   4, signed),
                               self.s2n(offset+4, 4, signed))
          values.append(value_j)
          offset = offset + typelen
      # Now "values" is either a string or an array
      a.append((tag,type,values))
    return a

def decode_ifd (fields, dict=EXIF_TAGS):
  result = []
  for (tag,type,values) in fields:
    try:
      stag = dict[tag]
    except:
      stag = '0x%04X' % tag
    stype = ['B',	# BYTE
             'A',	# ASCII
             'S',	# SHORT
             'L',	# LONG
             'R',	# RATIONAL
             'SB',	# SBYTE
             'U',	# UNDEFINED
             'SS',	# SSHORT
             'SL',	# SLONG
             'SR',	# SRATIONAL
            ] [type-1]
    result.append ((stag, values))
  return result

def get_marker (file):
	position = file.tell()
	marker = file.read (2)
	size = file.read (2)
	size = (ord(size[0])*256 + ord(size[1]))
	# size includes the two 'data-size' bytes
	return position, marker, file.read (size-2)

def process_file(file):
  d = []

  head = file.read(2)
  if head != '\377\330':
    # FFD8 = SOI = start of image
    raise ValueError, 'Not a JPEG file'
  else:
    while 1:
      # search for Exif/APP1 marker (0xffe1)
      position, marker, data = get_marker (file)
      #print 'marker=%s length=%d' % (repr(marker), len(data))
      if marker == '\377\341':
        # skip over the next six bytes, should be "Exif\000\000"
        assert (data[:6] == "Exif\000\000")
        data = data[6:]
        break

    d.append ({'I':'Intel', 'M':'Motorola'}[data[0]] + ' format')
    T = TIFF_file(data)
    L = T.list_IFDs()
    for i in range(len(L)):
      ifd = T.dump_IFD (L[i])
      r = decode_ifd (ifd)
      for tag, type, values in ifd:
        if tag == 0x8769:
          sub_ifd = T.dump_IFD (values[0])
          sub_r = decode_ifd (sub_ifd)
          for tag, type, values in sub_ifd:
            if tag == 0xa005:
              sub_sub_ifd = T.dump_IFD (values[0])
              sub_sub_r = decode_ifd (sub_sub_ifd)
              sub_r.append (sub_sub_r)
          r.append (sub_r)
      name = ''
      if i == 0:
        name = 'main image'
      elif i == 1:
        name = 'thumbnail'
      else:
        name = '???'
      d.append ((name, r))
    return d

def main():
  if len(sys.argv) < 2:
    sys.stderr.write('Usage: %s files...\n' % sys.argv[0])
    sys.exit(2)
  for filename in sys.argv[1:]:
    print filename+':'
    try:
      file = open(filename, 'rb')
      process_file(file)
    except IOError:
      print ' Cannot open file'
  sys.exit(0)

if __name__ == '__main__':
  main()
